// stack_vec_poin_off.hh, simple stack and vector implementations,
// poin_t struct, offsets_t struct.
/****************************************************************************

    C-formula interpreter library forint
    Copyright (C) 2000 Alexey Filin

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
    02111-1307, USA.
              
 ****************************************************************************/

#include <stdlib.h>
#include <iostream>
#include <string.h>

#include "forint_types.hh"

#ifndef _STACK_VEC_POIN_OFF_HH_
#define _STACK_VEC_POIN_OFF_HH_

namespace FInt {

/*****************************************************************************
*/

template <class T>
class stack_t
{
  T*   ptr;
  uint lim;
  uint last;
  const uint step;
public:
stack_t() : step(FI_REALLOC_SIZE) { ptr = 0; lim = 0; last = 0; }
stack_t(const stack_t& t) : step(FI_REALLOC_SIZE)
{
  ptr  = 0;
  lim  = t.lim;
  last = t.last;
  if (t.ptr) {
    ptr = (T*)malloc(lim*sizeof(T));
    memcpy(ptr,t.ptr,(last+1)*sizeof(T));
  }
}
stack_t&
operator = (const stack_t& t)
{
  if (this != &t) {
    if (ptr && !t.ptr) {
      free(ptr);
      ptr = 0;
    }
    last = t.last;
    if (t.ptr) {
      if (t.lim > lim)
	ptr = (T*)realloc(ptr,(lim = t.lim)*sizeof(T));
      memcpy(ptr,t.ptr,((last = t.last)+1)*sizeof(T));
    }
  }
  return *this;
}
virtual
~stack_t() { if (ptr) free(ptr); }

void
del(void) { if (ptr) free(ptr); ptr = 0; lim = 0; last = 0; }

void
cut(void) { last=0; }

// to get fast access to elements...
T*
getPtr() const { return ptr; }

uint
size() const { return (lim) ? last + 1 : 0; }

void
push(const T& t)
{
  if (!lim) {
    ptr = (T*)malloc((lim = step)*sizeof(T));
    *ptr = t;
  } else {
    if (lim == ++last)
      ptr = (T*)realloc(ptr, (lim += step)*sizeof(T));
    *(ptr + last) = t;
  }
}

T&
pop() { return *(ptr + last--); } // there is no error handling at all !!!

// insert 't' before an element in position 'pos'
bool
insert(uint pos, const T& t)
{
  if (pos>last) push(t);
  else {
    if (!lim) {
      ptr = (T*)malloc((lim = step)*sizeof(T));
      *ptr = t;
      return true;
    }
    else if (lim == ++last) ptr = (T*)realloc(ptr, (lim += step)*sizeof(T));
    T* p = ptr+pos;
    T tmp1;
    T tmp2 = t;
    uint s = last - pos + 1;
    while (s--) {
      tmp1 = *p;
      *p++ = tmp2;
      tmp2 = tmp1;
    }
  }
  return pos > last;
}

// delete an element in position 'pos', return 0 if successful
bool
erase(uint pos)
{
  if (pos>last || lim == 0) return false;
  if (last==0) {
    lim = 0;
    free(ptr);
  }
  else {
    uint s = last - pos;
    T* p = ptr+pos;
    while (s--) {
      *p = *(p+1);
      ++p;
    }
    --last;
  }
  return true;
}

// replace an element in position 'pos' by t, return 0 if successful
bool
replace(uint pos, const T& t)
{
  if (last < pos) return true;
  *(ptr + pos) = t;
  return false;
}

// for debugging
void show(void);
};

/*****************************************************************************
*/

// simple vector...
template <class T>
class vec_t
{
  T*   ptr;
  uint lim;
  uint last;
  const uint step;
public:
vec_t() : step(FI_REALLOC_SIZE) { ptr = 0; lim = 0; last = 0; }
vec_t(const vec_t& t) : step(FI_REALLOC_SIZE)
{
  ptr  = 0;
  lim  = t.lim;
  last = t.last;
  if (t.ptr) {
    ptr = (T*)malloc(lim*sizeof(T));
    memcpy(ptr,t.ptr,(last+1)*sizeof(T));
  }
}
vec_t&
operator = (const vec_t& t)
{
  if (this != &t) {
    if (ptr && !t.ptr) {
      free(ptr);
      ptr = 0;
    }
    last = t.last;
    if (t.ptr) {
      if (t.lim > lim)
	ptr = (T*)realloc(ptr,(lim = t.lim)*sizeof(T));
      memcpy(ptr,t.ptr,((last = t.last)+1)*sizeof(T));
    }
  }
  return *this;
}
virtual
~vec_t() { if (ptr) free(ptr); }

void
del(void) { if (ptr) free(ptr); ptr = 0; lim = 0; last = 0; }

void
cut(void) { last=0; }

// to get fast access to elements...
T*
getPtr() const { return ptr; }

uint
size() const { return (lim) ? last + 1 : 0; }

void
push_back(const T& t)
{
  if (!lim) {
    ptr = (T*)malloc((lim = step)*sizeof(T));
    *ptr = t;
  } else {
    if (lim == ++last)
      ptr = (T*)realloc(ptr, (lim += step)*sizeof(T));
    *(ptr + last) = t;
  }
}
T&
pop_back() { return *(ptr + last--); } // there is no error handling at all !!!

// insert 't' before an element in position 'pos'
bool
insert(uint pos, const T& t)
{
  if (pos>last) push_back(t);
  else {
    if (!lim) {
      ptr = (T*)malloc((lim = step)*sizeof(T));
      *ptr = t;
      return true;
    }
    else if (lim == ++last) ptr = (T*)realloc(ptr, (lim += step)*sizeof(T));
    T* p = ptr+pos;
    T tmp1;
    T tmp2 = t;
    uint s = last - pos + 1;
    while (s--) {
      tmp1 = *p;
      *p++ = tmp2;
      tmp2 = tmp1;
    }
  }
  return pos > last;
}

// delete an element in position 'pos', return 0 if successful
bool
erase(uint pos)
{
  if (pos>last || lim == 0) return false;
  if (last==0) {
    lim = 0;
    free(ptr);
  }
  else {
    uint s = last - pos;
    T* p = ptr+pos;
    while (s--) {
      *p = *(p+1);
      ++p;
    }
    --last;
  }
  return true;
}

// replace an element in position 'pos' by t, return 0 if successful
bool
replace(uint pos, const T& t)
{
  if (last < pos) return true;
  *(ptr + pos) = t;
  return false;
}

// for debugging
void show(void);
};

/*****************************************************************************
*/

struct poin_t
{
  const char* poin_type;
  b_type_t b_type;
  uint     level;
  uint  num_dims;
  uint* dim_size;
poin_t(void)
{
  poin_type= 0;
  b_type   = FI_CHAR;
  num_dims = 0;
  dim_size = 0;
}
virtual
~poin_t(void)
{
  if (dim_size)
    free(dim_size);
}

poin_t (const poin_t& t)
{
  dim_size = 0;

  poin_type = t.poin_type;
  b_type = t.b_type;
  level  = t.level;
  num_dims = t.num_dims;
  if (num_dims) {
    dim_size = (uint*)malloc(t.num_dims*sizeof(uint));
    memcpy(dim_size,t.dim_size,t.num_dims*sizeof(uint));
  }
}

poin_t&
operator = (const poin_t& t)
{
  if (this != &t) {
    poin_type = t.poin_type;
    b_type = t.b_type;
    level  = t.level;
    if (num_dims) {
      free(dim_size);
      dim_size = 0;
    }
    num_dims = t.num_dims;
    if (num_dims) {
      dim_size = (uint*)malloc(t.num_dims*sizeof(uint));
      memcpy(dim_size,t.dim_size,t.num_dims*sizeof(uint));
    }
  }
  return *this;
}
};

/*****************************************************************************
*/
struct offsets_t
{
  int ra;
  int sa;
  int num;
  int var;
  int poin;
  int poin_level;
  int func0;
  int func1;
  int func2d;
  int func2p;
  int func2;
  int offsets;

offsets_t(void)
{
  ra = 0;
  sa = 0;
  num = 0;
  var = 0;
  poin = 0;
  poin_level = 0;
  func0 = 0;
  func1 = 0;
  func2d = 0;
  func2p = 0;
  func2 = 0;
  offsets = 0;
}
virtual
~offsets_t(void) {}

};

}

#endif // _STACK_VEC_POIN_OFF_HH_
